#include "MenuHandlesMap.h"
#include "Window.h"
#include "Internals.h"
#include "Handlers_stub.h"
#include <stdlib.h>

MenuHandlesMap *newMenuHandlesMap()
{
	MenuHandlesMap *pMap;

	pMap = (MenuHandlesMap *) malloc(sizeof(MenuHandlesMap));
	if (!pMap)
		return NULL;

	pMap->children    = NULL;
	pMap->popups      = NULL;
	pMap->pFreeList   = NULL;
	pMap->pBlocks     = NULL;
	pMap->nAccelCount = 0;
	pMap->hAccelTable = NULL;
	pMap->nMenuID     = 0;

	memset(pMap->HashTable, 0, sizeof(MenuHandle)*HASH_TABLE_SIZE);

	return pMap;
}

void deleteMenuHandlesMap(MenuHandlesMap *pMap)
{
	Block *p, *pNext;	

	pMap->pFreeList = NULL;

	p = pMap->pBlocks;
	while (p != NULL)
	{
		pNext = p->pNext;
		free(p);
		p = pNext;
	}

	if (pMap->hAccelTable)
		DestroyAcceleratorTable(pMap->hAccelTable);

	free(pMap);
}

MenuHandle newMenuHandle(MenuHandlesMap *pMap, MenuHandle parent, MENU_TYPE type, int pos)
{
	int i;
	MenuHandle handle, child, *prev;
	unsigned int nHash;

	// it doesn't exist, add a new Association
	if (pMap->pFreeList == NULL)
	{
		// add another block
		Block *p = (Block *) malloc(sizeof(Block) + HASH_BLOCK_SIZE * sizeof(*handle));
		if (!p) return 0;
		p->pNext = pMap->pBlocks;
		pMap->pBlocks = p;  // change head (adds in reverse order for simplicity)

		// chain them into free list
		handle = (MenuHandle) (p+1);

		// free in reverse order to make it easier to debug
		handle += HASH_BLOCK_SIZE - 1;
		for (i = HASH_BLOCK_SIZE-1; i >= 0; i--, handle--)
		{
			handle->next = pMap->pFreeList;
			pMap->pFreeList = handle;
		}
	}

	handle = pMap->pFreeList;
	pMap->pFreeList = pMap->pFreeList->next;

	handle->parent   = parent;
	handle->id       = ++pMap->nMenuID;
	handle->key      = 0;
	handle->mods     = 0;
	handle->bitmap   = NULL;
	handle->children = NULL;
	handle->type     = type;
	handle->hMenu    = NULL;
	handle->nextInGroup = handle;
	handle->sibling  = NULL;
	
	if (type != MENU_POPUP)
	{
		if (parent)
		{
			child =  parent->children;
			prev  = &parent->children;
		}
		else
		{
			child =  pMap->children;
			prev  = &pMap->children;
		}
	}
	else
	{
		child =  pMap->popups;
		prev  = &pMap->popups;
	}

	if (pos < 0)
		while (child)
		{
			prev  = &child->sibling;
			child =  child->sibling;
		}
	else
		while (child && pos > 0)
		{
			prev  = &child->sibling;
			child =  child->sibling;			
			pos--;
		}
	handle->sibling = child;
	*prev = handle;

	// put into hash table
	nHash = (handle->id >> 4) % HASH_TABLE_SIZE;
	handle->next = pMap->HashTable[nHash];
	pMap->HashTable[nHash] = handle;

	return handle;
}

void deleteMenuHandle(MenuHandlesMap *pMap, MenuHandle handle)
{
	int nHash;
	MenuHandle child, *prev;
	
	while (handle->children)
		deleteMenuHandle(pMap, handle->children);
	
	if (handle->type != MENU_POPUP)
	{
		if (handle->parent)
		{
			prev  = &handle->parent->children;
			child = handle->parent->children;
		}
		else
		{
			prev  = &pMap->children;
			child = pMap->children;
		}
	}
	else
	{
		prev  = &pMap->popups;
		child = pMap->popups;
	}

	while (child)
	{
		if (child == handle)
		{
			*prev = handle->sibling;
			break;
		}

		prev  = &child->sibling;
		child = child->sibling;
	}
	
	nHash = (handle->id >> 4) % HASH_TABLE_SIZE;
	prev  = &(pMap->HashTable[nHash]);
	child = pMap->HashTable[nHash];
	
	while (child)
	{
		if (child == handle)
		{
			*prev = handle->next;
			break;
		}

		prev  = &child->next;
		child = child->next;
	}
	
	child = handle->nextInGroup;
	while (child->nextInGroup != handle)
		child = child->nextInGroup;
	child->nextInGroup = handle->nextInGroup;
	
	handle->next = pMap->pFreeList;
	pMap->pFreeList = handle;
}

MenuHandle getMenuHandle(MenuHandlesMap *pMap, UINT id)
{
	MenuHandle handle;
	unsigned int nHash;

	nHash = (id >> 4) % HASH_TABLE_SIZE;

	for (handle = pMap->HashTable[nHash]; handle != NULL; handle = handle->next)
	{
		if (handle->id == id)
			return handle;
	}

	return NULL;
}

void updateAccelTable(MenuHandlesMap *pMap, MenuHandle handle, int  key, unsigned int  mods)
{
	if ( handle->key && !key) pMap->nAccelCount--;
	if (!handle->key &&  key) pMap->nAccelCount++;
		
	handle->key  = key;
	handle->mods = mods;
	
	if (pMap->hAccelTable)
	{
		DestroyAcceleratorTable(pMap->hAccelTable);
		pMap->hAccelTable = NULL;
	}
};

HACCEL getAccelTableFromMap(MenuHandlesMap *pMap)
{
	int i,k;
	MenuHandle handle;
	ACCEL *pAccel;
	
	if (pMap->hAccelTable)
		return pMap->hAccelTable;

	if (pMap->nAccelCount == 0)
		return NULL;
	
	pAccel = malloc(sizeof(ACCEL)*pMap->nAccelCount);
	if (!pAccel)
		return NULL;

	for (k = 0, i = 0; k < pMap->nAccelCount; i++)
	{
		handle = pMap->HashTable[i];
		while (handle)
		{
			pAccel[k].cmd   = handle->id;
			
			if (handle->key > 0 && handle->key < 256)
			{
				pAccel[k].key   = handle->key;
				pAccel[k].fVirt = 0;
				k++;
			}
			else
				if (handle->key > 256 && handle->key < 512)
				{
					pAccel[k].key   = handle->key-256;
					pAccel[k].fVirt = FALT;
					k++;
				}
				else
				{					
				 	pAccel[k].fVirt = (((handle->mods & shiftBIT) ? FSHIFT   : 0) |
				                       ((handle->mods & ctrlBIT ) ? FCONTROL : 0) |
				            		   ((handle->mods & altBIT  ) ? FALT     : 0) | FVIRTKEY);
				            		   
				    switch (handle->key)
				    {
				    case kbUp:       pAccel[k++].key = VK_UP;     break;
					case kbDown:     pAccel[k++].key = VK_DOWN;   break;
					case kbLeft:     pAccel[k++].key = VK_LEFT;   break;
					case kbRight:    pAccel[k++].key = VK_RIGHT;  break;
					case kbPgUp:     pAccel[k++].key = VK_PRIOR;  break;
					case kbPgDown:   pAccel[k++].key = VK_NEXT;   break;
					case kbEnd:      pAccel[k++].key = VK_END;    break;
					case kbBegin:    pAccel[k++].key = VK_HOME;   break;
					case kbBackSpace:pAccel[k++].key = VK_BACK;   break;
					case kbDelete:   pAccel[k++].key = VK_DELETE; break;
					case kbEnter:    pAccel[k++].key = VK_RETURN; break;
					case kbEscape:   pAccel[k++].key = VK_ESCAPE; break;
					case kbTab:      pAccel[k++].key = VK_TAB;    break;
					case kbHelp:     pAccel[k++].key = VK_HELP;   break;
					case kbF1:       pAccel[k++].key = VK_F1;     break;
					case kbF2:       pAccel[k++].key = VK_F2;     break;
					case kbF3:       pAccel[k++].key = VK_F3;     break;
					case kbF4:       pAccel[k++].key = VK_F4;     break;
					case kbF5:       pAccel[k++].key = VK_F5;     break;
					case kbF6:       pAccel[k++].key = VK_F6;     break;
					case kbF7:       pAccel[k++].key = VK_F7;     break;
					case kbF8:       pAccel[k++].key = VK_F8;     break;
					case kbF9:       pAccel[k++].key = VK_F9;     break;
					case kbF10:      pAccel[k++].key = VK_F10;    break;
					case kbF11:      pAccel[k++].key = VK_F11;    break;
					case kbF12:      pAccel[k++].key = VK_F12;    break;
					case kbClear:    pAccel[k++].key = VK_CLEAR;  break;
					}
				}
			
			handle = handle->next;
		}
	}
	
	pMap->hAccelTable = CreateAcceleratorTable(pAccel, pMap->nAccelCount);

	free(pAccel);
	
	return pMap->hAccelTable;
};

int getMenuPos(MenuHandlesMap *pMap, MenuHandle handle)
{
	int pos;
	MenuHandle h;
	
	if (handle->type != MENU_POPUP)
		h = handle->parent ? handle->parent->children : pMap->children;
	else
		h = pMap->popups;

	pos = 0;
	while (h && h != handle)
	{
		pos++;
		h = h->sibling;
	}

	return pos;
};

HMENU getParentHMENU(MenuHandle handle)
{
	if (handle->type != MENU_POPUP)
		if (handle->parent == NULL)
			return GetMenu(ghWndFrame);
		else
			return handle->parent->hMenu;
	else
		return NULL;
}

int getChildrenCount(MenuHandlesMap *pMap, MenuHandle handle)
{
	int count;
	MenuHandle child;
	
	count = 0;
	child = handle ? handle->children : pMap->children;
	while (child != NULL)
	{
		count++;
		child = child->sibling;
	}
	
	return count;
}

MenuHandle getNthChild(MenuHandlesMap *pMap, MenuHandle handle, int index)
{
	MenuHandle child;

	child = handle ? handle->children : pMap->children;
	while (child != NULL && index > 0)
	{
		index--;
		child = child->sibling;
	}
		
	return child;
}

void notifyHandleForDestroy(MenuHandle handle)
{
    MenuHandle child;
    child = handle->children;
    while (child)
    {
       notifyHandleForDestroy(child);
       child = child->sibling;
    }
	handleMenuDestroy(handle);
}

void notifyAllHandlesForDestroy(MenuHandlesMap *pMap)
{
	MenuHandle child;
	
    child = pMap->children;
    while (child)
    {
       notifyHandleForDestroy(child);
       child = child->sibling;
    }
    
    child = pMap->popups;
	while (child)
	{
	   notifyHandleForDestroy(child);
	   child = child->sibling;
    }
}
